import time as tm
import numpy as np
import datetime
from pyomo.environ import *
from pyomo.opt import SolverFactory
from openpyxl import load_workbook
import csv
import matplotlib.pyplot as plt
import dill as pickle


data = np.loadtxt(open('returns.csv'), skiprows=1, delimiter=',')
parames_raw = np.loadtxt(open('parameters/parameters_t.txt'), skiprows=1, delimiter='\t')

load_sims = True
Res = []

for cf in np.linspace(0.01,0.1,19):
#for cf in [0.09]:
    mp = parames_raw[:,0] # mortality probabilities
    print(cf)
    num_sim = 100
    num_lags = 5
    T = 35
    num_rows, num_assets = data.shape
    sigma = 1
    _lambda = 100
    cl = 0.90 # confidence level
    A = 0.05
    #cf = 0.09
    infl = 0.03
    
    # penalty function
    #kappa = np.zeros([T])
    #
    #for t in range(T):
    #    temp = 0
    #    for j in range(t,T):
    #        temp += mp[j]*(j-t+1)*10
    #    kappa[t] = temp
    kappa = 2*np.power(1.3, np.linspace(35,1,35))
    
    
    disc_f = list(map(lambda x: (1+infl)**(-x), range(1,36)))
    
    R = np.zeros([num_assets, T, num_sim])
    
    if load_sims == False:
        for s in range(num_sim):
            rand_idx = np.random.randint(0,num_rows,[T,1])
            for t in range(T):
                R[:,t,s] = data[rand_idx[t],:]-0.12
    
#        np.save(open('returns2.obj','wb'),R)
    else:
#        R = np.loadtxt('parameters/parameters_r_its.txt',skiprows=1).T.reshape((num_assets,T,num_sim),order='F')
        R=np.load(open('returns2_stress.obj','rb'))
    
    K = np.zeros([T, num_sim, num_sim])
    
    # creating a composite return matrix, firs num_lag returns come from the historical data, rest is the simulation
    Rk = np.zeros([num_assets,T+num_lags,num_sim])
    for i in range(num_assets):
        for t in range(T):
            for s in range(num_sim):
                Rk[i, t+num_lags-1, s] = R[i,t,s]
   
    
    for i in range(num_assets):
        for t in range(num_lags):
            for s in range(num_sim):
                Rk[i,t,s] = data[-t,i]
    
    # kernel matrix
    if True:
        print('creating Kernel matrixes')
        for t in range(T):
            print(t)
            for k in range(num_sim):
                for s in range(num_sim):
                    temp = 0
    #                for j in range(num_assets):
    #                    for l in range(t):
                    temp += sum(sum((Rk[:,t:t+num_lags,k]-Rk[:,t:t+num_lags,s])**2))/(num_lags)
                            
                    K[t,k,s] = np.exp(-temp/float(sigma))
                    
        print('done creating Kernel matrixes')                
    else:
        K = pickle.load(open('kernel2.pkl','rb'))
    
    
    
    
    model = ConcreteModel()
    # number of assets
    model.N = Set(initialize = range(num_assets))
    
    # time periods
    model.T = Set(initialize = range(T))
    
    # number of simulations
    model.S = Set(initialize = range(num_sim))
    
    model.u = Var(model.N, model.T, bounds = (None, None))
    model.y = Var(model.N, model.S, bounds = (None, None))
    #model.B = Var(model.T, model.S, bounds = (0, cf))
    model.z = Var(bounds = (0, None))
    model.zeta = Var(model.T, bounds = (None, None))
    model.x = Var(model.N, model.T, model.S, bounds = (0, None))
    model.x_init = Var(model.N, bounds = (0, None))
    model.V = Var(model.T, model.S, bounds = (0, None))
    model.temp = Var(model.T, model.S, bounds = (0, None)) # cvar linearization aux. variable
    model.h = Var(model.N, model.T, model.S, bounds = (-cf,0))
    model.cf_gap = Var(model.T, model.S, bounds = (0,None))
    model.t = Var(model.N, model.T, model.S, bounds = (0, None))
    V0 = 0.5
    
    
    def obj(model):
        alpha = -1/float(num_sim)*sum([sum([model.V[t,s]*mp[t]*disc_f[t] for t in model.T]) for s in model.S])
        beta = _lambda*sum([sum([model.y[i,s]**2 for s in model.S]) for i in model.N])
        gamma = sum([sum([model.cf_gap[t,s]*kappa[t] for t in model.T]) for s in model.S])
        return  alpha + beta + gamma
    
    def init_value(model,t,s):
        return model.V[t,s] == sum([model.x[i,t,s] for i in model.N])
            
    def investment_dynamics(model,i,t,s):
        if t >= 1:
            return model.x[i,t,s] == (1+R[i,t,s])*(model.x[i,t-1,s] + model.u[i,t] +\
                   sum([model.y[i,j]*K[t,j,s] for j in model.S ]) + model.h[i,t,s])
        else:
            return model.x[i,0,s] == (1+R[i,0,s])*model.x_init[i]
    

    def leverage_constraint(model,t,s):
        if t>=1:
            return sum([ model.t[i,t,s] for i in model.N]) <= 0.2*model.V[t-1,s]
        else:
            return sum([model.t[i,t,s] for i in model.N]) <= 0.2*(V0- model.z)


    def uit_is_zero(model,i,t):
        return model.u[i,t] == 0

    def abs_lin1(model,i,t,s):
        return model.t[i,t,s] >= model.u[i,t] + sum([model.y[i,j]*K[t,j,s] for j in model.S ])

    def abs_lin2(model, i, t, s):
        return model.t[i,t,s] >= -model.u[i,t] - sum([model.y[i,j]*K[t,j,s] for j in model.S ])

    def monotonicity(model, t, s):
        if t>=1:
            return sum([model.h[i,t-1,s] for i in model.N]) <=\
                   sum([model.h[i,t,s] for i in model.N])
        else:
            return Constraint.Feasible
        
    
    def CVaR_cash_outflow(model, t, s):
        return  model.cf_gap[t,s] >= sum([model.h[i,t,s] for i in model.N]) - A*model.z + cf
    
    
    #def CVaR_linearization(model, t, s):
    #    return model.temp[t,s] >= sum([model.h[i,t,s] for i in model.N]) - A*model.z - model.zeta[t]
    
    def adjustment_balance(model,t,s):
        return sum([model.u[i,t] + sum([model.y[i,j]*K[t,j,s] for j in model.S ]) for i in model.N]) == 0
    
    def initial_investment(model):
        return sum([model.x_init[i] for i in model.N])  == V0 - model.z
    
    
    #def investment_continuity(model,t,s):
    #    return sum([model]) == sum()
    
    model.prob_obj = Objective(rule = obj, sense = minimize)
    
    model.init_value = Constraint(model.T, model.S, rule = init_value)
    model.investment_dynamics = Constraint(
            model.N, model.T, model.S, 
            rule = investment_dynamics
    )
    model.monotonicity = Constraint(model.T, model.S, rule = monotonicity) 
    model.CVaR_cash_outflow = Constraint(model.T, model.S, rule = CVaR_cash_outflow) 
    model.leverage_constraint = Constraint(model.T, model.S, rule = leverage_constraint)
    model.abs_lin1 = Constraint(model.N, model.T, model.S, rule = abs_lin1)
    model.abs_lin2 = Constraint(model.N, model.T, model.S, rule = abs_lin2)
    #model.CVaR_linearization = Constraint(model.T, model.S, rule = CVaR_linearization) 
    model.adjustment_balance = Constraint(model.T, model.S, rule = adjustment_balance)
    model.initial_investment = Constraint(rule = initial_investment)
    model.uit_is_zero = Constraint(model.N, model.T, rule = uit_is_zero)
    
    print('starting optimization')
    t0 = tm.time()
    opt = SolverFactory('gurobi')
    results = opt.solve(model)
    print('solving time ', tm.time() - t0)
    
    X = np.zeros([num_assets, T, num_sim])
    
    for i in model.N:
        for t in model.T:
            for s in model.S:
                X[i,t,s] = model.x[i,t,s].value
    
    
    V = np.zeros([T, num_sim])
    
    for t in model.T:
        for s in model.S:
            V[t,s] = model.V[t,s].value
            
    average_value = np.average(V, axis=1)
    
    
    u_its = np.zeros([num_assets, T, num_sim])
    
    for i in model.N:
        for t in model.T:
            for s in model.S:
                temp = model.u[i,t].value + sum([model.y[i,j].value*K[t,j,s] for j in model.S ])
                u_its[i,t,s] = temp
            
            
            
    B = np.zeros([T, num_sim])
    for t in model.T:
        for s in model.S:
            B[t,s] = model.cf_gap[t,s].value 
            
            
    average_value = np.average(V, axis=1)
    
    cf_array = np.zeros([T, num_sim])
    for t in model.T:
        for s in model.S:
            cf_array[t,s] = sum([model.h[i,t,s].value for i in model.N])
            
    Res.append({'target_cf':cf, 
                'cf_array':cf_array, 
                'target_cf':cf,
                'V':V, 
                'B':B, 
                'model.h':model.h,
                'model_cf_gap':model.cf_gap,
                'model_u':model.u, 
                'model_z':model.z,
                'model_y':model.y,
                'model_x':model.x,
                'model_x_init':model.x_init})


pickle.dump(Res, open('res_stress.pkl','wb'))
#    
#    plt.figure(1)
#    plt.plot(cf[:,4])
#    #
#    plt.figure(2)
#    plt.plot(range(1,T+1), average_value, 'bo-')
#    plt.plot(range(1,T+1), V, 'k-', alpha = 0.1)
#    







