import time as tm
import numpy as np
import datetime
from pyomo.environ import *
from pyomo.opt import SolverFactory
from openpyxl import load_workbook
import csv
import matplotlib.pyplot as plt
import dill as pkl
import scipy as sp
import scipy.io as sio


scenario_type = '_stress'
#scenario_type = ''
markers = [
        'o-',
        '+-',
        'D-',
        's-',
        '*-',
        ]

V_test_avg_array = []
cf_array = []

model_all = pkl.load(open('res%s.pkl'%scenario_type,'rb'))

class Model(object):
    def __init__(self, i):
        model = model_all[i]
        self.z = model['model_z']
        self.x = model['model_x']
        self.y = model['model_y']
        self.cf = model['cf_array']
        self.x_init = model['model_x_init']
        self.V = model['V']
        self.B = model['B']
        self.N = range(10)
        self.T = range(35)
        self.S = range(100)
        self.target_cf = model['target_cf']
        
        
plt_cfs = []

allocation_by_cf={}

for midx in range(0,19,4):
    model = Model(midx)
    data = np.loadtxt(open('returns.csv'), skiprows=1, delimiter=',')
    parames_raw = np.loadtxt(open('parameters/parameters_t.txt'), skiprows=1, delimiter='\t')
    R=np.load(open('returns2%s.obj'%scenario_type,'rb'))
    
    N = 10
    S = 100
    T = 35
    num_lags = 5
    num_assets = N
    num_sim = S
    cf = model.target_cf
    V0 = 0.5-model.z.value
    sigma = 1    
    annuity = model.z.value

    # creating a composite return matrix, firs num_lag returns come from the historical data, rest is the simulation
    Rk = np.zeros([num_assets,T+num_lags,num_sim])
    for i in range(num_assets):
        for t in range(T):
            for s in range(num_sim):
                Rk[i, t+num_lags, s] = R[i,t,s]
       
    
    for i in range(num_assets):
        for t in range(num_lags):
            for s in range(num_sim):
                Rk[i,t,s] = data[-t,i]
    
    
    # getting individual investments
    X = np.zeros([N,T,S])
    
    for i in model.N:
        for t in model.T:
            for s in model.S:
                X[i,t,s] = model.x[i,t,s].value
    
    avgX = np.mean(X, axis = 2)
    
    
    X0 = np.zeros(N)
    for i in model.N:
        X0[i] = model.x_init[i].value   
    
    U0 = np.zeros([N,T])
    
    # getting kernel parameters
    Y = np.zeros([N,T,S])
    
    for i in model.N:
        for t in model.T:
            for s in model.S:
                Y[i,t,s] = model.y[i,s].value
    
    
    # loading the scenarios
    r_test_temp = np.load(open('scenarios/returns_test%s.obj'%scenario_type,'rb'))
    returns = Rk
    
    
    r_test = np.zeros([num_assets,T+num_lags,num_sim])
    for i in range(num_assets):
        for t in range(T):
            for s in range(num_sim):
                r_test[i, t+num_lags, s] = r_test_temp[i,t,s]
       
    
    for i in range(num_assets):
        for t in range(num_lags):
            for s in range(num_sim):
                r_test[i,t,s] = data[-t,i]
                

    def rbf(r, N, t, k, gamma=1):
        w = 0
        if(t>0):
            temp =  (r - returns[:,t:t+num_lags,k])**2
            temp2 = np.sum(temp, axis = 0)/(num_lags)        
            w = np.sum(temp)
    
        return np.exp(-gamma*w)
    

    def adj_f(r, i, t):
        w = 0
        for k in range(S):
            w = w + Y[i,t,k] * rbf(r, N, t, k ) 
        return w + U0[i,t]
    

    u_its = np.zeros([N,T,S])
    for i in range(N):
        for s in range(S): 
            print(i,s)
            for t in range(35):
                u_its[i,t,s] = (adj_f(r_test[:,t:t+num_lags,s], i, t))
    

    
    V_test = np.zeros([T,S])
    V0_test = sum(X0)
    X_test = np.zeros([N,T,S])
    
    for t in range(T):
        for s in range(S):
            for i in range(N):
                if t==0:
                    X_test[i,t,s] = X0[i]*(1+r_test[i,t+num_lags,s]) 
                elif sum([max(X_test[i,t-1,s],0) for i in model.N])>0 :
                    X_test[i,t,s] = (
                            X_test[i,t-1,s]+u_its[i,t,s] - \
                            cf*max(X_test[i,t-1,s],0)/sum([max(X_test[i,t-1,s],0) for i in model.N]))*(1+r_test[i,t+num_lags,s]) 
                else:
                    X_test[i,t,s] = 0
        
            V_test[t,s] = sum(X_test[:,t,s])
    
    
    
    X_test_max = np.zeros([N,T,S])
    
    for t in range(T):
        for s in range(S):
            S_minus = 0
            S_plus = 0
            
            for i in range(N):
                if X_test[i,t,s] < 0:
                    S_minus -= X_test[i,t,s]         
                else:
                    S_plus +=X_test[i,t,s]
                    
            
            for i in range(N):
                if X_test[i,t,s] < 0:
                    X_test_max[i,t,s] = 0
                else:
                    X_test_max[i,t,s] = X_test[i,t,s] - S_minus* X_test[i,t,s]/S_plus if S_minus<S_plus else X_test[i,t,s]
    
    X_test_avg = np.mean(X_test_max, axis = 2)
    allocation_by_cf[cf] = np.hstack((np.matrix(X0).T,X_test_avg))
    np.savetxt(open('output%s/X_avg.txt'%scenario_type,'wb'), X_test_avg)
    
    V_test_max_sum = np.sum(X_test_max, axis=0)
    
    V_test_avg = np.mean(V_test_max_sum, axis = 1)
    np.savetxt(open('output%s/V_avg.txt'%scenario_type,'wb'), X_test_avg)
    
    
    
    
    V_test_static = np.ones([T,S])/2.0
    for s in range(S):
        for t in range(1,T):
            V_test_static[t,s] = np.max(V_test_static[t-1,s]*(1 + np.average(r_test[:,t-1+num_lags,s])) - cf, 0 )
    
    V_test_static_avg = np.mean(V_test_static, axis=1)
    
    V_test_avg_array.append(V_test_avg)
    cf_array.append(cf)
    
for v in V_test_avg_array:
    plt.plot(range(1,36), v ,markers.pop(),
             color='black', lw=1,
             markersize=5, 
             fillstyle='full', 
             markerfacecolor='lightgray'
    )
plt.xlabel('Years')
plt.ylabel('Average Portfolio Value (in millions of dollars)')
plt.legend(['L = '+str(int(z*1000))+'K' for z in cf_array])
plt.savefig('output%s/V_test.eps'%scenario_type)
plt.savefig('output%s/V_test.png'%scenario_type)










